from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Optional, Sequence, Union

import numpy as np
import sympy as sp
from smitfit.typing import Numerical


@dataclass
class Parameter:
    """A mutable parameter class that supports method chaining"""

    symbol: sp.Symbol
    guess: Numerical = 1.0
    bounds: tuple[Optional[Numerical], Optional[Numerical]] = (None, None)
    fixed: bool = False

    @property
    def name(self) -> str:
        return self.symbol.name

    @property
    def shape(self) -> tuple[int, ...]:
        shape = getattr(self.guess, "shape", tuple())
        return shape

    def fix(self) -> Parameter:
        """Fix the parameter at its current guess value"""
        self.fixed = True
        return self

    def unfix(self) -> Parameter:
        """Make the parameter free to vary"""
        self.fixed = False
        return self

    def set_bounds(
        self, lower: Optional[Numerical] = None, upper: Optional[Numerical] = None
    ) -> Parameter:
        """Set parameter bounds"""
        self.bounds = (lower, upper)
        return self

    def set_guess(self, value: Numerical) -> Parameter:
        """Set initial guess value"""
        self.guess = value
        return self


class Parameters:
    """Container for managing multiple parameters"""

    def __init__(self, parameters: Iterable[Parameter]):
        self._parameters = {p.name: p for p in parameters}

    def __getitem__(self, key: Union[str, Sequence[str]]) -> Union[Parameter, Parameters]:
        """Get a parameter or create a new ParameterSet with selected parameters"""
        if isinstance(key, str):
            return self._parameters[key]
        return Parameters([self._parameters[k] for k in key])

    def __iter__(self):
        return iter(self._parameters.values())

    def __len__(self):
        return len(self._parameters)

    def __add__(self, other: Parameters) -> Parameters:
        return Parameters(self.to_list() + other.to_list())

    @classmethod
    def from_symbols(
        cls,
        symbols: Iterable[sp.Symbol],
    ) -> Parameters:
        symbol_dict = {symbol.name: symbol for symbol in sorted(symbols, key=str)}

        p_list = [Parameter(symbol) for symbol in symbol_dict.values()]
        return cls(p_list)

    @property
    def guess(self) -> dict[str, Numerical]:  # other types?
        return {p.name: np.asarray(p.guess) for p in self}

    def fix(self, *names: str) -> Parameters:
        """Fix specified parameters"""
        for name in names:
            self._parameters[name].fix()
        return self

    def unfix(self, *names: str) -> Parameters:
        """Unfix specified parameters"""
        for name in names:
            self._parameters[name].unfix()
        return self

    def set_bounds(
        self, bounds_dict: dict[str, tuple[Optional[Numerical], Optional[Numerical]]]
    ) -> Parameters:
        """Set bounds for multiple parameters at once"""
        for name, (lower, upper) in bounds_dict.items():
            self._parameters[name].set_bounds(lower, upper)
        return self

    def set_guesses(self, guess_dict: dict[str, Numerical]) -> Parameters:
        """Set initial guesses for multiple parameters at once"""
        for name, guess in guess_dict.items():
            self._parameters[name].set_guess(guess)
        return self

    @property
    def fixed(self) -> Parameters:
        """Get list of fixed parameters"""
        return Parameters([p for p in self._parameters.values() if p.fixed])

    @property
    def free(self) -> Parameters:
        """Get list of free parameters"""
        return Parameters([p for p in self._parameters.values() if not p.fixed])

    def to_list(self) -> list[Parameter]:
        """Convert to parameter list"""
        return list(self._parameters.values())

    def __repr__(self) -> str:
        return f"Parameters({list(self._parameters.values())})"


def unpack(x: np.ndarray, shapes: dict[str, tuple[int, ...]]) -> dict[str, Numerical]:
    """Unpack a ndim 1 array of concatenated parameter values into a dictionary of
    parameter name: parameter_value where parameter values are cast back to their
    specified shapes.
    """
    sizes = [int(np.prod(shape)) for shape in shapes.values()]

    x_split = np.split(x, np.cumsum(sizes))
    p_values = {name: arr.reshape(shape) for (name, shape), arr in zip(shapes.items(), x_split)}

    return p_values


def pack(
    parameter_values: Iterable[np.ndarray],
) -> np.ndarray:  # todo iterable numerical dtype input
    """Pack a dictionary of parameter_name together as array"""

    return np.concatenate(tuple(param_value.ravel() for param_value in parameter_values))



# %%

import numpy as np
from smitfit.loss import MSELoss
from smitfit.minimize import Minimize
from smitfit.model import Model
from smitfit.symbol import Symbols

# %%

s = Symbols("x y a b")
model = Model({s.y: s.a * s.x + s.b})  # type: ignore

# Generate Ground-Truth data
np.random.seed(43)
gt = {"a": 0.15, "b": 2.5}

xdata = np.linspace(0, 11, num=100)
ydata = gt["a"] * xdata + gt["b"]

noise = np.random.normal(0, scale=ydata / 10.0 + 0.2)
ydata += noise
# %%
parameters = model.define_parameters("a b")
parameters
# %%

loss = MSELoss(model, dict(y=ydata))
objective = Minimize(loss, parameters, dict(x=xdata))
result = objective.fit()
result

# %%
# %%
# compare to numpy polyfit
np.polyfit(xdata, ydata, deg=1)


# %%
import ultraplot as uplt

fig, ax = uplt.subplots()
ax.scatter(xdata, ydata)
ax.plot(xdata, model(**result.parameters, x=xdata)["y"], color="r")

# %%



"""
Same problem as in 'markov_chain_analytical', but now solved numerically rather than with
matrix exponentiation.

The numerical IVP solver is useful for larger models where matrix exponentiation is no longer
feasible.

"""

from __future__ import annotations

import numpy as np
import ultraplot as uplt
from smitfit.composite_expr import MarkovIVP
from smitfit.loss import MSELoss
from smitfit.markov import extract_states, generate_transition_matrix
from smitfit.minimize import Minimize
from smitfit.model import Model
from smitfit.parameter import Parameters
from smitfit.symbol import symbol_matrix
from sympy import Symbol

# %%
np.random.seed(43)

# %%
# Ground truth parameter values for generating data and fitting
gt_values = {
    "k_A_B": 1e0,
    "k_B_A": 5e-2,
    "k_B_C": 5e-1,
    "y0_A": 1.0,
    "y0_B": 0.0,
    "y0_C": 0.0,
}

# %%
# Generate markov chain transition rate matrix from state connectivity string(s)
connectivity = ["A <-> B -> C"]
m = generate_transition_matrix(connectivity)


# %%
states = extract_states(connectivity)
y0 = symbol_matrix(name="y0", shape=(3, 1), suffix=states)
model = Model({Symbol("y"): MarkovIVP(Symbol("t"), m, y0)})

rate_params = Parameters.from_symbols(m.free_symbols)
y0_params = Parameters.from_symbols(y0.free_symbols)

parameters = rate_params + y0_params
# %%
num = 50
xdata = {"t": np.linspace(0, 11, num=num)}

# Calling a matrix based model expands the dimensions of the matrix on the first axis to
# match the shape of input variables or parameters.
populations = model(**gt_values, **xdata)["y"]
populations.shape

# %%
# add noise to populations
ydata = {"y": populations + np.random.normal(0, 0.05, size=num * 3).reshape(populations.shape)}
ydata["y"].shape  # shape of the data is (3, 50)

# %%
loss = MSELoss(model, ydata)
objective = Minimize(loss, parameters, xdata)
result = objective.fit()

# %%
# Compare fit result with ground truth parameters
for k, v in result.parameters.items():
    print(f"{k:5}: {v:10.2}, ({gt_values[k]:10.2})")

# %%
color = ["#7FACFA", "#FA654D", "#8CAD36"]
cycle = uplt.Cycle(color=color)

eval_data = {"t": np.linspace(0, 11, 1000)}
y_eval = model(**result.parameters, **eval_data)["y"]

fig, ax = uplt.subplots()
ax.scatter(xdata["t"], ydata["y"].T, cycle=cycle)
ax.line(eval_data["t"], y_eval.T, cycle=cycle)
ax.format(xlabel="Time", ylabel="Population Fraction")
uplt.show()



from __future__ import annotations

from functools import reduce
from typing import Any, Iterable, Optional, OrderedDict, Union

import numpy as np
from smitfit.parameter import Parameter


def flat_concat(data: dict[str, np.ndarray]) -> np.ndarray:
    return np.concatenate([arr.flatten() for arr in data.values()])


def clean_types(d: Any) -> Any:
    """cleans up nested dict/list/tuple/other `d` for exporting as yaml

    Converts library specific types to python native types, including numpy dtypes,
    OrderedDict, numpy arrays

    # https://stackoverflow.com/questions/59605943/python-convert-types-in-deeply-nested-dictionary-or-array

    """
    if isinstance(d, np.floating):
        return float(d)

    if isinstance(d, np.integer):
        return int(d)

    if isinstance(d, np.ndarray):
        return d.tolist()

    if isinstance(d, list):
        return [clean_types(item) for item in d]

    if isinstance(d, tuple):
        return tuple(clean_types(item) for item in d)

    if isinstance(d, OrderedDict):
        return clean_types(dict(d))

    if isinstance(d, dict):
        return {k: clean_types(v) for k, v in d.items()}

    else:
        return d


def format_indexer(indexer: tuple[slice, int, None, Ellipsis]) -> str:
    """Format a tuple of slice objects into a string that can be used to index a numpy array.

    More or less the inverse of `numpy.index_exp`.


    Args:
        indexer: Tuple of indexing objects.

    """

    return f"[{', '.join(_format_indexer(sl) for sl in indexer)}]"


def _format_indexer(indexer: Union[slice, int, None, Ellipsis]) -> str:
    if isinstance(indexer, int):
        return str(indexer)
    elif isinstance(indexer, slice):
        # adapted from
        # https://stackoverflow.com/questions/24662999/how-do-i-convert-a-slice-object-to-a-string-that-can-go-in-brackets
        sl_start = "" if indexer.start is None else str(indexer.start)
        sl_stop = "" if indexer.stop is None else str(indexer.stop)
        if indexer.step is None:
            sl_str = "%s:%s" % (sl_start, sl_stop)
        else:
            sl_str = "%s:%s:%s" % (sl_start, sl_stop, indexer.step)
        return sl_str
    elif isinstance(indexer, type(None)):
        return "None"
    elif isinstance(indexer, type(Ellipsis)):
        return "..."
    else:
        raise TypeError(f"Unexpected type: {type(indexer)}")


# https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties/31174427#31174427
def rsetattr(obj: Any, attr: str, val: Any) -> Any:
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


# https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties/31174427#31174427
def rgetattr(obj: Any, attr: str, *default):
    try:
        return reduce(getattr, attr.split("."), obj)
    except AttributeError as e:
        if default:
            return default[0]
        else:
            raise e



import numpy as np
import numpy.typing as npt
import sympy as sp
from smitfit.utils import Optional


class Symbols(dict):
    def __init__(self, names, cls=sp.Symbol, **kwargs):
        default_kwargs = {"seq": True}
        default_kwargs.update(kwargs)
        super().__init__({s.name: s for s in sp.symbols(names, cls=cls, **default_kwargs)})

    def __repr__(self):
        return f"Symbols({list(self.keys())})"

    def __getattr__(self, name) -> sp.Symbol:
        if name in self:
            return self[name]
        raise AttributeError(f"'SymbolNamespace' object has no attribute '{name}'")


def symbol_matrix(
    name: Optional[str] = None,
    shape: Optional[tuple[int, ...]] = None,
    names: Optional[npt.ArrayLike] = None,
    suffix: Optional[npt.ArrayLike] = None,
) -> sp.Matrix:
    if shape is None:
        if names is not None:
            shape = (len(names), 1)
        elif suffix is not None:
            shape = (len(suffix), 1)
        else:
            raise ValueError("If 'shape' is not given, must specify 'names' or 'suffix'")

    # Generate names for parameters. Uses 'names' first, then <name>_<suffix> otherwise generates suffices
    # from indices
    if names is None and name is None:
        raise ValueError("Must specify either 'name' or 'names'")
    elif names is None:
        names = np.full(shape, fill_value="", dtype=object)
        if suffix is None:
            for i, j in np.ndindex(shape):
                names[i, j] = f"{name}_{i}_{j}"
        else:
            suffix = np.array(suffix).reshape(shape)
            for i, j in np.ndindex(shape):
                names[i, j] = f"{name}_{suffix[i, j]}"
    else:
        names = np.array(names)

    matrix = sp.zeros(*shape)
    for i, j in np.ndindex(shape):
        matrix[i, j] = sp.Symbol(
            name=names[i, j],
        )

    return matrix



# %%
from __future__ import annotations

from functools import cached_property
from typing import Callable, Union

import numpy as np
import sympy as sp
from smitfit.typing import Numerical


# %%
class Expr:
    def __init__(self, expr) -> None:
        self._expr = expr

    @property
    def expr(self):
        return self._expr

    @cached_property
    def symbols(self) -> set[sp.Symbol]:
        return set()

    def filter_kwargs(self, **kwargs) -> dict[str, Numerical]:
        """Parse kwargs and take only the ones in `free_parameters`"""
        try:
            kwargs = {k: kwargs[k] for k in {s.name for s in self.symbols}}
        except KeyError as e:
            raise KeyError(f"Missing value for {e}") from e

        return kwargs

    def __getitem__(self, item):
        return GetItem(self, item)

    def __call__(self, **kwargs):
        return self._expr


class GetItem(Expr):
    def __init__(self, expr: Expr, item: Union[tuple, slice, int]):
        # todo super
        self._expr = expr
        self.item = item

    def __call__(self, **kwargs):
        ans = self._expr(**kwargs)
        return ans[self.item]

    @cached_property
    def symbols(self) -> set[sp.Symbol]:
        return self._expr.symbols

    def __repr__(self) -> str:
        return f"{self.expr.__repr__()}[{self.item!r}]"


class SympyExpr(Expr):
    @cached_property
    def symbols(self) -> set[sp.Symbol]:
        return self._expr.free_symbols

    @cached_property
    def lambdified(self) -> Callable:
        ld = sp.lambdify(sorted(self.symbols, key=str), self._expr)

        return ld

    def __repr__(self) -> str:
        return f"SympyExpr({self._expr})"

    def __call__(self, **kwargs):
        return self.lambdified(**self.filter_kwargs(**kwargs))


class SympyMatrixExpr(Expr):
    def __init__(self, expr: sp.MatrixBase) -> None:
        super().__init__(expr)

    @cached_property
    def symbols(self) -> set[sp.Symbol]:
        return self._expr.free_symbols

    @cached_property
    def lambdified(self) -> dict[tuple[int, int], Callable]:
        lambdas = {}
        for i, j in np.ndindex(self.expr.shape):
            lambdas[(i, j)] = sp.lambdify(sorted(self.symbols, key=str), self.expr[i, j])

        return lambdas

    def __call__(self, **kwargs):
        # when marix elements != scalars, shape is expanded by the first dimensions to accomodate.
        ld_kwargs = self.filter_kwargs(**kwargs)

        base_shape = np.broadcast_shapes(
            *(getattr(value, "shape", tuple()) for value in ld_kwargs.values())
        )

        # squeeze last dim if shape is (1,)
        base_shape = () if base_shape == (1,) else base_shape
        shape = base_shape + self.expr.shape

        out = np.empty(shape)
        for i, j in np.ndindex(self.expr.shape):
            out[..., i, j] = self.lambdified[i, j](**ld_kwargs)

        return out


def as_expr(expr) -> Expr | dict[str, Expr]:
    if isinstance(expr, Expr):
        return expr
    elif isinstance(expr, (float, np.ndarray)):  # torch tensor, ...
        return Expr(expr)
    elif isinstance(expr, sp.MatrixBase):
        return SympyMatrixExpr(expr)
    if isinstance(expr, sp.Expr):
        return SympyExpr(expr)
    elif isinstance(expr, dict):
        return {k: as_expr(v) for k, v in expr.items()}  # type: ignore
    else:
        raise TypeError(f"Invalid type: {type(expr)}")



from __future__ import annotations

import os
import pickle
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np
from smitfit.utils import clean_types, rgetattr


@dataclass
class FitResult:
    """
    Fit result object.
    """

    fit_parameters: dict[str, np.ndarray]
    """Fitted parameter values"""

    gof_qualifiers: dict
    """Goodness-of-fit qualifiers"""

    errors: dict[str, float | np.ndarray] = field(default_factory=dict)

    fixed_parameters: dict[str, float | np.ndarray] = field(default_factory=dict)  # Numerical dtype
    """Values of the model's fixed parameters"""

    guess: Optional[dict] = None
    """Initial guesses"""

    minimizer: Optional[Minimizer] = None

    hessian: Optional[np.ndarray] = None

    metadata: dict = field(default_factory=dict)
    """Additional metadata"""

    base_result: Optional[Any] = field(default=None, repr=False)
    """Source fit result object. Can be dicts of sub results"""

    def __post_init__(self) -> None:
        if "datetime" not in self.metadata:
            now = datetime.now()
            self.metadata["datetime"] = now.strftime("%Y/%m/%d %H:%M:%S")
            self.metadata["timestamp"] = int(now.timestamp())

    def __str__(self):
        if any(np.ndim(v) != 0 for v in self.fit_parameters.values()):
            raise ValueError("Cannot print fit result with array values.")

        s = ""
        # stdev = self.stdev if self.hessian is not None else None

        p_size = max(len(k) for k in self.fit_parameters)
        if self.errors:
            s += f"{'Parameter':<{p_size}} {'Value':>10} {'Stdev':>10}\n"
        else:
            s += f"{'Parameter':<{p_size}} {'Value':>10}\n"

        for k, v in self.fit_parameters.items():
            s += f"{k:<{max(p_size, 9)}} {v:>10.3g}"
            if self.errors:
                s += f" {self.errors[k]:>10.3g}"
            s += "\n"

        return s

    def to_dict(self) -> dict:
        """
        Convert the fit result to a dictionary.

        Returns:
            Dictionary representation of the fit result.
        """
        keys = [
            "gof_qualifiers",
            "fit_parameters",
            "fixed_parameters",
            "guess",
            "metadata",
        ]
        if self.hessian is not None:
            keys += ["stdev"]

        d = {k: v for k in keys if (v := getattr(self, k)) is not None}

        return clean_types(d)

    def to_yaml(self, path: Union[os.PathLike[str], str], sort_keys: bool = False) -> None:
        """
        Save the fit result as yaml.

        Args:
            path: Path to save to.
            sort_keys: Boolean indicating whether to sort the keys.

        """
        import yaml

        Path(path).write_text(yaml.dump(self.to_dict(), sort_keys=sort_keys))

    def to_pickle(self, path: Union[os.PathLike[str], str]) -> None:
        """
        Save the fit result as pickle.

        Args:
            path: Path to save to.
        """
        try:
            del self.minimizer.model.numerical
        except AttributeError:
            pass

        with Path(path).open("wb") as f:
            pickle.dump(self, f)

    def eval_hessian(self, hessian: Optional[Hessian] = None) -> np.ndarray:
        # TODO Hessian as parameter / strategy
        """Evaluate the hessian at the fitted parameters values"""

        hessian = hessian or rgetattr(self.minimizer, "objective.hessian", None)
        if hessian is None:
            raise ValueError("No Hessian available")

        if hasattr(self.minimizer, "objective") and list(hessian.shapes.items()) != list(
            self.minimizer.objective.shapes.items()
        ):
            raise ValueError("Mismatch between objective and fit parameters")

        x = pack(self.fit_parameters.values())
        self.hessian = hessian(x)

        return self.hessian

    @property
    def variance(self) -> dict[str, float | np.ndarray]:
        if self.hessian is None:
            self.eval_hessian()

        hess_inv = np.linalg.inv(self.hessian)
        var = np.diag(hess_inv)
        parameter_shapes = {k: v.shape for k, v in self.fit_parameters.items()}
        return unpack(var, parameter_shapes)

    @property
    def stdev(self) -> dict[str, float | np.ndarray]:
        return {k: np.sqrt(v) for k, v in self.variance.items()}

    @property
    def parameters(self) -> dict[str, float | np.ndarray]:
        return {**self.fit_parameters, **self.fixed_parameters}



from __future__ import annotations

from functools import cached_property
from typing import Optional

import numpy as np
import sympy as sp
from scipy.integrate import solve_ivp
from smitfit.expr import Expr, as_expr


class CompositeExpr(Expr):
    @cached_property
    def symbols(self) -> set[sp.Symbol]:
        return set().union(*(expr.symbols for expr in self.expr.values()))

    def __call__(self, **kwargs):
        return {k: v(**kwargs) for k, v in self.expr.items()}


class MarkovIVP(CompositeExpr):
    """Uses scipy.integrate.solve_ivp to numerically find time evolution of a markov process
        given a transition rate matrix.

    Returned shape is <states>, <datapoints>

    """

    def __init__(
        self,
        t: sp.Symbol | sp.Expr | Expr,
        trs_matrix: sp.Matrix | Expr,
        y0: sp.Matrix | Expr,
        domain: Optional[tuple[float, float]] = None,
        **ivp_kwargs,
    ):
        expr = as_expr({"t": t, "trs_matrix": trs_matrix, "y0": y0})
        super().__init__(expr)

        ivp_defaults = {"method": "Radau"}
        self.ivp_defaults = ivp_defaults | ivp_kwargs
        self.domain = domain

    def __call__(self, **kwargs):
        components = super().__call__(**kwargs)

        # if `self['t']` does not depend on any parameters; domain can be precomputed and
        # does not have to be determined for every call
        # although its every fast to do so
        domain = self.domain or self.get_domain(components["t"])
        sol = solve_ivp(
            self.grad_func,
            domain,
            y0=components["y0"].squeeze(),
            t_eval=components["t"],
            args=(components["trs_matrix"],),
            **self.ivp_defaults,
        )

        return sol.y

    def get_domain(self, arr: np.ndarray) -> tuple[float, float]:
        # padding?
        return arr[0], arr[-1]

    @staticmethod
    def grad_func(t, y, trs_matrix):
        return trs_matrix @ y



from __future__ import annotations

import re
from typing import Iterable

import sympy as sp
from smitfit.expr import as_expr
from smitfit.parameter import Parameter, Parameters
from smitfit.typing import Numerical
from toposort import toposort


class Model:
    def __init__(self, model: dict) -> None:
        self.model = model
        self.expr: dict = {k: as_expr(v) for k, v in self.model.items()}
        topology = {k: v.symbols for k, v in self.expr.items()}
        self.call_stack = [
            elem for subset in toposort(topology) for elem in subset if elem in model.keys()
        ]

    @property
    def x_symbols(self) -> set[sp.Symbol]:
        return set.union(*(v.symbols for v in self.expr.values())) - self.y_symbols

    @property
    def y_symbols(self) -> set[sp.Symbol]:
        return set(self.model.keys())

    def __call__(self, **kwargs):
        resolved = {}
        for key in self.call_stack:
            resolved[key.name] = self.expr[key](**kwargs, **resolved)
        return resolved

    def define_parameters(
        self, parameters: dict[str, Numerical] | Iterable[str] | str = "*"
    ) -> Parameters:
        symbols = {s.name: s for s in self.x_symbols}
        if parameters == "*":
            params = [Parameter(symbol) for symbol in self.x_symbols]
        elif isinstance(parameters, str):
            params = [Parameter(symbols[k]) for k in re.split(r"; |, |\*|\s+", parameters)]
        elif isinstance(parameters, dict):
            params = [Parameter(symbols[k], guess=v) for k, v in parameters.items()]
        else:
            raise TypeError("Invalid type")

        return Parameters(params)



import numpy as np
from smitfit.loss import MSELoss
from smitfit.minimize import Minimize
from smitfit.model import Model
from smitfit.symbol import Symbols


def test_fit():
    s = Symbols("x y a b")
    model = Model({s.y: s.a * s.x + s.b})  # type: ignore

    # Generate Ground-Truth data
    np.random.seed(43)
    gt = {"a": 0.15, "b": 2.5}

    xdata = np.linspace(0, 11, num=100)
    ydata = gt["a"] * xdata + gt["b"]

    noise = np.random.normal(0, scale=ydata / 10.0 + 0.2)
    ydata += noise
    parameters = model.define_parameters("a b")

    loss = MSELoss(model, dict(y=ydata))
    objective = Minimize(loss, parameters, dict(x=xdata))
    result = objective.fit()

    expected = {"a": np.array(0.14567421), "b": np.array(2.5566292)}
    for k, v in result.parameters.items():
        assert np.allclose(v, expected[k])



from typing import Optional

from smitfit.model import Model
from smitfit.reduce import mean_reduction, sum_reduction


class Loss:
    """sum/average reduction"""

    def __call__(self, **kwargs) -> float:
        return 0.0


class SELoss(Loss):
    """sum of squared errors"""

    def __init__(self, model: Model, y_data: dict, weights: Optional[dict] = None):
        self.model = model
        self.y_data = y_data
        self.weights = weights or {}

    def squares(self, **kwargs):
        y_model = self.model(**kwargs)

        squares = {
            k: ((y_model[k] - self.y_data[k]) * self.weights.get(k, 1)) ** 2
            for k in self.y_data.keys()
        }

        return squares

    def __call__(self, **kwargs) -> float:
        squares = self.squares(**kwargs)

        return sum_reduction(squares)


class MSELoss(SELoss):
    def __call__(self, **kwargs) -> float:
        squares = self.squares(**kwargs)

        return mean_reduction(squares)



from __future__ import annotations

import numpy as np
from scipy.optimize import minimize
from smitfit.fitresult import FitResult
from smitfit.loss import Loss, SELoss
from smitfit.parameter import Parameters, pack, unpack
from smitfit.utils import flat_concat


class Minimize:  # = currently only scipy minimize
    def __init__(self, loss: Loss, parameters: Parameters, xdata: dict[str, np.ndarray]):
        self.loss = loss
        self.parameters = parameters
        self.xdata = xdata
        self.shapes = {p.name: p.shape for p in self.parameters}

    def func(self, x: np.ndarray):
        parameters = unpack(x, self.shapes)
        return self.loss(**parameters, **self.xdata)

    @property
    def free_parameters(self):
        return [p for p in self.parameters if not p.fixed]

    @property
    def fixed_parameters(self):
        return [p for p in self.parameters if p.fixed]

    def get_bounds(self) -> list[tuple[float | None, float | None]] | None:
        bounds = []
        for p in self.free_parameters:
            size = np.prod(p.shape, dtype=int)
            bounds += [p.bounds] * size

        if all((None, None) == b for b in bounds):
            return None
        else:
            return bounds

    def fit(self):
        x = pack(self.parameters.free.guess.values())
        result = minimize(self.func, x, bounds=self.get_bounds())
        fit_parameters = unpack(result.x, self.shapes)

        gof_qualifiers = {
            "loss": result["fun"],
        }

        std_error = {}
        if hasattr(self.loss, "y_data"):
            y_data = self.loss.y_data
            ans = self.loss.model(**self.xdata, **fit_parameters)
            f = flat_concat({k: ans[k] for k in y_data})
            y = flat_concat(y_data)

            gof_qualifiers["r_squared"] = 1 - np.sum((y - f) ** 2) / np.sum((y - np.mean(y)) ** 2)

            # move this to a function somewhere
            if isinstance(self.loss, SELoss):
                residuals = y - f

                N = len(y)
                P = len(fit_parameters)
                s_squared = np.sum(residuals**2) / (N - P)

                hess_inv = result.hess_inv  # or calculate one with numdiff
                cov_mat = s_squared * hess_inv * 2
                std_error_arr = np.sqrt(np.diag(cov_mat))
                std_error = unpack(std_error_arr, self.shapes)

        return FitResult(
            fit_parameters=fit_parameters,
            gof_qualifiers=gof_qualifiers,
            errors=std_error,
            fixed_parameters=self.parameters.fixed.guess,
            guess=self.parameters.free.guess,
            base_result=result,
        )
